#!/bin/sh

if test "${KERNEL_VER-set}" = set; then echo "KERNEL_VER is not set"; exit 1; fi
if test "${TARGETNAME-set}" = set; then echo "TARGETNAME is not set"; exit 1; fi

PWD=`pwd`

case ${KERNEL_VER} in
	2.4 )
	OBJ=o
	MODVER=`modinfo -f%{kernel_version} ${PWD}/${TARGETNAME}.${OBJ}`
	;;
	2.6 )
	OBJ=ko
	MODVER=`modinfo -F vermagic ${PWD}/${TARGETNAME}.${OBJ} | cut -d' ' -f1`
	;;
esac

if test "${MODVER}" = "" ; then
	echo "Can not get kernel version from ${TARGETNAME}.${OBJ}."
	exit 1
fi

if test "${MODVER}" = "`uname -r`"; then
	echo "You made a module which is for current kernel ${MODVER}."
	echo "Deleting previous installed driver module ${TARGETNAME}..."
	HPTMV6=$( find /lib/modules/`uname -r`/* -name ${TARGETNAME}.${OBJ} )
	rm -f ${HPTMV6}
	rm -rf /lib/modules/`uname -r`/kernel/drivers/scsi/${TARGETNAME}
	mkdir -p /lib/modules/`uname -r`/kernel/drivers/scsi/${TARGETNAME}
	echo "Install the new driver module..."
	cp ${PWD}/${TARGETNAME}.${OBJ} /lib/modules/`uname -r`/kernel/drivers/scsi/${TARGETNAME}
	echo -n "Updating module dependencies..."
	depmod -a
	echo "Done."
elif [ -d /lib/modules/${MODVER} ]; then
	echo "You made a module for ${MODVER} which does not match current kernel."
	echo "The driver will be installed for kernel ${MODVER}."
	echo "Deleting previous installed driver module ${TARGETNAME}..."
	HPTMV6=$( find /lib/modules/${MODVER}/* -name ${TARGETNAME}.${OBJ} )
	rm -f ${HPTMV6}
	rm -rf /lib/modules/${MODVER}/kernel/drivers/scsi/${TARGETNAME}
	mkdir -p /lib/modules/${MODVER}/kernel/drivers/scsi/${TARGETNAME}
	echo "Install the new driver module..."
	cp -f ${PWD}/${TARGETNAME}.${OBJ} /lib/modules/${MODVER}/kernel/drivers/scsi/${TARGETNAME}
	echo -n "Updating module dependencies..."
	depmod -a ${MODVER}
	echo "Done."
else
	echo "The driver is built for kernel ${MODVER} and can not be installed to current system."
	exit 1
fi

echo "Checking for initrd images to be updated..."

which mkinitrd 1> /dev/null 2> /dev/null
if test $? -ne 0 ; then
	echo "Can not find command 'mkinitrd'."
	exit 1
fi

#### Next if the system is on controller, make the system loadable
# mandrake redhat fedora debian rhel turbo sles redflag
distinfo() {
	local issue=/etc/issue
	[ -e /etc/mandrake-release ] && { echo mandrake; return; }
	[ -e /etc/redhat-release ] && { echo redhat; return; }
	[ -e /etc/fedora-release ] && { echo fedora; return; }
	[ -e /etc/SuSE-release ] && { echo suse; return; }
	[ -e /etc/debian_version ] && { echo debian; return; }
	[ -e /etc/turbolinux-release ] && { echo turbo; return; }
	if [ -e $issue ]; then
		grep -i "debian" $issue && { echo debian; return; }
		grep -i "red *hat" $issue && { echo redhat; return; }
		grep -i "suse\|s\.u\.s\.e" $issue && { echo suse; return; }
	fi
}

DIST=`distinfo`
case ${DIST} in
mandrake | fedora | redhat )
	if [ -f /boot/initrd-${MODVER}.img ] ; then
		mkdir -p /tmp/${TARGETNAME}-loop
		gzip -dc /boot/initrd-${MODVER}.img > /tmp/${TARGETNAME}-loop.img 2>/dev/null
		mount -o loop /tmp/${TARGETNAME}-loop.img /tmp/${TARGETNAME}-loop 1>/dev/null 2>/dev/null
		if test $? != 0 ; then
			cpio -it < /tmp/${TARGETNAME}-loop.img | grep -q ${TARGETNAME}.${OBJ}
			if test $? = 0 ; then
				grep -q scsi_mod /lib/modules/${MODVER}/modules.dep
				if test $? = 0 ; then
					WITHSCSI=--preload\ scsi_mod
				fi
				grep -q sd_mod /lib/modules/${MODVER}/modules.dep
				if test $? = 0 ; then
					WITHSD=--preload\ scsi_mod
				fi
				if [ ! -f /boot/initrd-${MODVER}.img.${TARGETNAME} ]; then
					echo "Backup /boot/initrd-${MODVER}.img to /boot/initrd-${MODVER}.img.${TARGETNAME}."
					mv /boot/initrd-${MODVER}.img /boot/initrd-${MODVER}.img.${TARGETNAME}
				fi
				mkinitrd ${WITHSCSI} ${WITHSD} --with=${TARGETNAME} /boot/initrd-${MODVER}.img ${MODVER}
			fi
		else
			HPTMV6=$( find /tmp/${TARGETNAME}-loop/* -name ${TARGETNAME}.${OBJ} )
			if test "${HPTMV6}" != "" ; then
				cp -f ${PWD}/${TARGETNAME}.${OBJ} ${HPTMV6}
				umount /tmp/${TARGETNAME}-loop.img
				if [ ! -f /boot/initrd-${MODVER}.img.${TARGETNAME} ]; then
					echo "Backup /boot/initrd-${MODVER}.img to /boot/initrd-${MODVER}.img.${TARGETNAME}."
					mv /boot/initrd-${MODVER}.img /boot/initrd-${MODVER}.img.${TARGETNAME}
				fi
				gzip -9c /tmp/${TARGETNAME}-loop.img > /boot/initrd-${MODVER}.img
			else
				echo "There is no need to update the initrd image."
				umount /tmp/${TARGETNAME}-loop.img
			fi
		fi
		rmdir /tmp/${TARGETNAME}-loop 2>/dev/null
		rm -f /tmp/${TARGETNAME}-loop.img
	fi
;;
suse  )
	if test ${MODVER} = `uname -r` ; then
	 	if [ -f /boot/initrd-${MODVER} ] ; then
			if [ ! -f /boot/initrd-${MODVER}.${TARGETNAME} ]; then
				echo "Backup /boot/initrd-${MODVER} to /boot/initrd-${MODVER}.${TARGETNAME}."
				mv /boot/initrd-${MODVER} /boot/initrd-${MODVER}.${TARGETNAME}
			fi
		fi
		mkinitrd
	else
		echo "The compiled module is not for current kernel. /boot/initrd-`uname -r` is not updated."
	fi
;;
debian )
	if [ -f /boot/initrd.img-${MODVER} ] ; then
		if [ ! -f /boot/initrd.img-${MODVER}.${TARGETNAME} ]; then
			echo "backup /boot/initrd.img-${MODVER} to /boot/initrd.img-${MODVER}.${TARGETNAME}."
			mv /boot/initrd.img-${MODVER} /boot/initrd.img-${MODVER}.${TARGETNAME}
		fi
		mkinitrd -o /boot/initrd.img-${MODVER} ${MODVER}
	fi	
;;
turbo )
	cd /boot
	if [ -f /boot/initrd-${MODVER} ] ; then
		mkdir -p /tmp/${TARGETNAME}-loop
		gzip -dc /boot/initrd-${MODVER} > /tmp/${TARGETNAME}-loop.img 2>/dev/null
		mount -o loop /tmp/${TARGETNAME}-loop.img /tmp/${TARGETNAME}-loop
		HPTMV6=$( find /tmp/${TARGETNAME}-loop/* -name ${TARGETNAME}.${OBJ} )
		if test "${HPTMV6}" != "" ; then
			cp -f ${PWD}/${TARGETNAME}.${OBJ} ${HPTMV6}
			umount /tmp/${TARGETNAME}-loop.img
			if [ ! -f /boot/initrd-${MODVER}.${TARGETNAME} ]; then
				echo "Backup /boot/initrd-${MODVER} to /boot/initrd-${MODVER}.${TARGETNAME}."
				mv /boot/initrd-${MODVER} /boot/initrd-${MODVER}.${TARGETNAME}
			fi
			gzip -9c /tmp/${TARGETNAME}-loop.img > /boot/initrd-${MODVER}
		else
			echo "There is no need to update the initrd image."
			umount /tmp/${TARGETNAME}-loop.img
		fi
		rmdir /tmp/${TARGETNAME}-loop 2>/dev/null
		rm -f /tmp/${TARGETNAME}-loop.img
	fi
	;;
* )
	echo "Unknown distribution - You have to update the initrd image manually."
;;
esac
